"""
TopasOpt is a library for performing inverse optimisation with Topas Monte Carlo.
Please see the docs page for more info:
https://acrf-image-x-institute.github.io/TopasOpt/
"""
__author__ = 'Brendan Whelan'
__version__ = '0.3.0'


